#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from jdmn.feel.lib.BaseFEELLibTest import BaseFEELLibTest
from jdmn.feel.lib.Types import STRING
from jdmn.runtime.Context import Context
from jdmn.runtime.LambdaExpression import LambdaExpression
from jdmn.runtime.Range import Range


class BaseStandardFEELLibTest(BaseFEELLibTest):
    """
    Base test class for BaseStandardFEELLib
    """
    __test__ = False

    #
    # Conversion functions
    #
    def testNumberWithSeparators(self):
        self.assertIsNone(self.getLib().number(None, None, None))
        self.assertIsNone(self.getLib().number(None, ".", ","))
        self.assertIsNone(self.getLib().number("1.235.00", ".", "."))
        self.assertIsNone(self.getLib().number("12,356,00", ".", ","))

        self.assertEqualsNumber(self.makeNumber("12356"), self.getLib().number("12.356,00", ".", ","))
        self.assertEqualsNumber(self.makeNumber("12356"), self.getLib().number("12.356", ".", ","))
        self.assertEqualsNumber(self.makeNumber("12356.01"), self.getLib().number("12356,01", ".", ","))

        self.assertEqualsNumber(self.makeNumber("12356"), self.getLib().number("12,356.00", ",", "."))
        self.assertEqualsNumber(self.makeNumber("12356"), self.getLib().number("12,356", ",", "."))
        self.assertEqualsNumber(self.makeNumber("12356.01"), self.getLib().number("12356.01", ",", "."))

        self.assertEqualsNumber(self.makeNumber("1000000.01"), self.getLib().number("1000000.01", None, "."))
        self.assertEqualsNumber(self.makeNumber("1000000.00"), self.getLib().number("1,000,000", ",", None))
        self.assertEqualsNumber(self.makeNumber("1000000.00"), self.getLib().number("1,000,000.00", ",", None))
        self.assertEqualsNumber(self.makeNumber("1000000.01"), self.getLib().number("1.000.000,01", ".", ","))

    def testDuration(self):
        self.assertEqualsDateTime("P1Y8M", self.getLib().duration("P1Y8M"))
        self.assertEqualsDateTime("P2DT20H", self.getLib().duration("P2DT20H"))
        self.assertEqualsDateTime("-PT2H", self.getLib().duration("-PT2H"))

        self.assertEqualsDateTime("P1Y8M", self.getLib().duration("P1Y8M"))
        self.assertEqualsDateTime("P2DT20H", self.getLib().duration("P2DT20H"))

        self.assertEqualsDateTime("P999999999M", self.getLib().duration("P999999999M"))
        self.assertEqualsDateTime("-P999999999M", self.getLib().duration("-P999999999M"))
        self.assertEqualsDateTime("P1Y0M2DT6H58M59.000S", self.getLib().duration("P1Y0M2DT6H58M59.000S"))
        # Overflow in duration(from)
        self.assertEqualsDateTime("P11999999988M", self.getLib().duration("P11999999988M"))
        self.assertEqualsDateTime("P2129706043D", self.getLib().duration("P2129706043D"))

    def testYearsAndMonthsDuration(self):
        self.assertIsNone(self.getLib().yearsAndMonthsDuration(None, None))

        self.assertEqualsDateTime("P0Y0M", self.getLib().yearsAndMonthsDuration(self.makeDate("2015-12-24"), self.makeDate("2015-12-24")))
        self.assertEqualsDateTime("P1Y2M", self.getLib().yearsAndMonthsDuration(self.makeDate("2016-09-30"), self.makeDate("2017-12-28")))
        self.assertEqualsDateTime("P7Y6M", self.getLib().yearsAndMonthsDuration(self.makeDate("2010-05-30"), self.makeDate("2017-12-15")))
        self.assertEqualsDateTime("-P4033Y2M", self.getLib().yearsAndMonthsDuration(self.makeDate("2014-12-31"), self.makeDate("-2019-10-01")))
        self.assertEqualsDateTime("-P4035Y11M", self.getLib().yearsAndMonthsDuration(self.makeDate("2017-09-05"), self.makeDate("-2019-10-01")))

    #
    # Numeric functions
    #
    def testDecimal(self):
        self.assertIsNone(self.getLib().decimal(None, None))
        self.assertIsNone(self.getLib().decimal(None, self.makeNumber("128")))
        self.assertIsNone(self.getLib().decimal(self.makeNumber("10"), None))

        #        self.assertEqualsNumber(self.makeNumber("-10"), self.getLib().decimal(self.makeNumber("-10"), self.makeNumber(Long.MAX_VALUE)))

        self.assertEqualsNumber(self.makeNumber("0.33"), self.getLib().decimal(self.makeNumber("0.333"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().decimal(self.makeNumber("1.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().decimal(self.makeNumber("2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("10.00"), self.getLib().decimal(self.makeNumber("10.001"), self.makeNumber("2")))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("10.00"), self.getLib().decimal(self.makeNumber("12.001"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("0E+2"), self.getLib().decimal(self.makeNumber("12.001"), self.makeNumber("-2")))
        self.assertEqualsNumber(self.makeNumber("0E+3"), self.getLib().decimal(self.makeNumber("12.001"), self.makeNumber("-3")))

    def testRoundWithUpMode(self):
        self.assertIsNone(self.getLib().round(None, None, None))
        self.assertIsNone(self.getLib().round(None, self.makeNumber("128"), None))
        self.assertIsNone(self.getLib().round(self.makeNumber("10"), self.makeNumber("2"), None))
        self.assertIsNone(self.getLib().round(self.makeNumber("10"), self.makeNumber("2"), "abc"))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().round(self.makeNumber("5.5"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().round(self.makeNumber("2.5"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("1.6"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("1.1"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.0"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.0"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-1.1"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-1.6"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("-3"), self.getLib().round(self.makeNumber("-2.5"), self.makeNumber("0"), "up"))
        self.assertEqualsNumber(self.makeNumber("-6"), self.getLib().round(self.makeNumber("-5.5"), self.makeNumber("0"), "up"))

        self.assertEqualsNumber(self.makeNumber("5.13"), self.getLib().round(self.makeNumber("5.125"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("2.13"), self.getLib().round(self.makeNumber("2.125"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().round(self.makeNumber("1.126"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().round(self.makeNumber("1.121"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.120"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.120"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().round(self.makeNumber("-1.121"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().round(self.makeNumber("-1.126"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("-2.13"), self.getLib().round(self.makeNumber("-2.125"), self.makeNumber("2"), "up"))
        self.assertEqualsNumber(self.makeNumber("-5.13"), self.getLib().round(self.makeNumber("-5.125"), self.makeNumber("2"), "up"))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-1"), "up"))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-1"), "up"))
        self.assertEqualsNumber(self.makeNumber("2E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-2"), "up"))
        self.assertEqualsNumber(self.makeNumber("-2E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-2"), "up"))

    def testRoundWithDownMode(self):
        self.assertIsNone(self.getLib().round(None, None, "down"))
        self.assertIsNone(self.getLib().round(None, self.makeNumber("128"), "down"))
        self.assertIsNone(self.getLib().round(self.makeNumber("10"), None, "down"))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("5"), self.getLib().round(self.makeNumber("5.5"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("2.5"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.6"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.1"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.0"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.0"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.1"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.6"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-2.5"), self.makeNumber("0"), "down"))
        self.assertEqualsNumber(self.makeNumber("-5"), self.getLib().round(self.makeNumber("-5.5"), self.makeNumber("0"), "down"))

        self.assertEqualsNumber(self.makeNumber("5.12"), self.getLib().round(self.makeNumber("5.125"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("2.12"), self.getLib().round(self.makeNumber("2.125"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.126"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.121"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.120"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.120"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.121"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.126"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("-2.12"), self.getLib().round(self.makeNumber("-2.125"), self.makeNumber("2"), "down"))
        self.assertEqualsNumber(self.makeNumber("-5.12"), self.getLib().round(self.makeNumber("-5.125"), self.makeNumber("2"), "down"))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.2E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-1"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1.2E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-1"), "down"))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-2"), "down"))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-2"), "down"))

    def testRoundWithHalfUpMode(self):
        self.assertIsNone(self.getLib().round(None, None, "half up"))
        self.assertIsNone(self.getLib().round(None, self.makeNumber("128"), "half up"))
        self.assertIsNone(self.getLib().round(self.makeNumber("10"), None, "half up"))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().round(self.makeNumber("5.5"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().round(self.makeNumber("2.5"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("1.6"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.1"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.0"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.0"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.1"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-1.6"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-3"), self.getLib().round(self.makeNumber("-2.5"), self.makeNumber("0"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-6"), self.getLib().round(self.makeNumber("-5.5"), self.makeNumber("0"), "half up"))

        self.assertEqualsNumber(self.makeNumber("5.13"), self.getLib().round(self.makeNumber("5.125"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("2.13"), self.getLib().round(self.makeNumber("2.125"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().round(self.makeNumber("1.126"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.121"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.120"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.120"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.121"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().round(self.makeNumber("-1.126"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-2.13"), self.getLib().round(self.makeNumber("-2.125"), self.makeNumber("2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-5.13"), self.getLib().round(self.makeNumber("-5.125"), self.makeNumber("2"), "half up"))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-1"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-1"), "half up"))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-2"), "half up"))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-2"), "half up"))

    def testRoundWithHalfDownMode(self):
        self.assertIsNone(self.getLib().round(None, None, "half down"))
        self.assertIsNone(self.getLib().round(None, self.makeNumber("128"), "half down"))
        self.assertIsNone(self.getLib().round(self.makeNumber("10"), None, "half down"))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("5"), self.getLib().round(self.makeNumber("5.5"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("2.5"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("1.6"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.1"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.0"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.0"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.1"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-1.6"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-2.5"), self.makeNumber("0"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-5"), self.getLib().round(self.makeNumber("-5.5"), self.makeNumber("0"), "half down"))

        self.assertEqualsNumber(self.makeNumber("5.12"), self.getLib().round(self.makeNumber("5.125"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("2.12"), self.getLib().round(self.makeNumber("2.125"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().round(self.makeNumber("1.126"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.121"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.120"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.120"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.121"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().round(self.makeNumber("-1.126"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-2.12"), self.getLib().round(self.makeNumber("-2.125"), self.makeNumber("2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-5.12"), self.getLib().round(self.makeNumber("-5.125"), self.makeNumber("2"), "half down"))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-1"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-1"), "half down"))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-2"), "half down"))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-2"), "half down"))

    def testRoundWithHalfEvenMode(self):
        self.assertIsNone(self.getLib().round(None, None, "half down"))
        self.assertIsNone(self.getLib().round(None, self.makeNumber("128"), "half down"))
        self.assertIsNone(self.getLib().round(self.makeNumber("10"), None, "half down"))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().round(self.makeNumber("5.5"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("2.5"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().round(self.makeNumber("1.6"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.1"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().round(self.makeNumber("1.0"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.0"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().round(self.makeNumber("-1.1"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-1.6"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().round(self.makeNumber("-2.5"), self.makeNumber("0"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-6"), self.getLib().round(self.makeNumber("-5.5"), self.makeNumber("0"), "half even"))

        self.assertEqualsNumber(self.makeNumber("5.12"), self.getLib().round(self.makeNumber("5.125"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("2.12"), self.getLib().round(self.makeNumber("2.125"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().round(self.makeNumber("1.126"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.121"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().round(self.makeNumber("1.120"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.120"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().round(self.makeNumber("-1.121"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().round(self.makeNumber("-1.126"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-2.12"), self.getLib().round(self.makeNumber("-2.125"), self.makeNumber("2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-5.12"), self.getLib().round(self.makeNumber("-5.125"), self.makeNumber("2"), "half even"))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-1"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-1"), "half even"))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().round(self.makeNumber("125.125"), self.makeNumber("-2"), "half even"))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().round(self.makeNumber("-125.125"), self.makeNumber("-2"), "half even"))

    def testRoundUp(self):
        self.assertIsNone(self.getLib().roundUp(None, None))
        self.assertIsNone(self.getLib().roundUp(None, self.makeNumber("128")))
        self.assertIsNone(self.getLib().roundUp(self.makeNumber("10"), None))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().roundUp(self.makeNumber("5.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().roundUp(self.makeNumber("2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().roundUp(self.makeNumber("1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().roundUp(self.makeNumber("1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundUp(self.makeNumber("1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundUp(self.makeNumber("-1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().roundUp(self.makeNumber("-1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().roundUp(self.makeNumber("-1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-3"), self.getLib().roundUp(self.makeNumber("-2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-6"), self.getLib().roundUp(self.makeNumber("-5.5"), self.makeNumber("0")))

        self.assertEqualsNumber(self.makeNumber("5.13"), self.getLib().roundUp(self.makeNumber("5.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("2.13"), self.getLib().roundUp(self.makeNumber("2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().roundUp(self.makeNumber("1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().roundUp(self.makeNumber("1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundUp(self.makeNumber("1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundUp(self.makeNumber("-1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().roundUp(self.makeNumber("-1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().roundUp(self.makeNumber("-1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-2.13"), self.getLib().roundUp(self.makeNumber("-2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-5.13"), self.getLib().roundUp(self.makeNumber("-5.125"), self.makeNumber("2")))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().roundUp(self.makeNumber("125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().roundUp(self.makeNumber("-125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("2E+2"), self.getLib().roundUp(self.makeNumber("125.125"), self.makeNumber("-2")))
        self.assertEqualsNumber(self.makeNumber("-2E+2"), self.getLib().roundUp(self.makeNumber("-125.125"), self.makeNumber("-2")))

    def testRoundDown(self):
        self.assertIsNone(self.getLib().roundDown(None, None))
        self.assertIsNone(self.getLib().roundDown(None, self.makeNumber("128")))
        self.assertIsNone(self.getLib().roundDown(self.makeNumber("10"), None))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("5"), self.getLib().roundDown(self.makeNumber("5.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().roundDown(self.makeNumber("2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundDown(self.makeNumber("1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundDown(self.makeNumber("1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundDown(self.makeNumber("1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundDown(self.makeNumber("-1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundDown(self.makeNumber("-1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundDown(self.makeNumber("-1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().roundDown(self.makeNumber("-2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-5"), self.getLib().roundDown(self.makeNumber("-5.5"), self.makeNumber("0")))

        self.assertEqualsNumber(self.makeNumber("5.12"), self.getLib().roundDown(self.makeNumber("5.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("2.12"), self.getLib().roundDown(self.makeNumber("2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundDown(self.makeNumber("1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundDown(self.makeNumber("1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundDown(self.makeNumber("1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundDown(self.makeNumber("-1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundDown(self.makeNumber("-1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundDown(self.makeNumber("-1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-2.12"), self.getLib().roundDown(self.makeNumber("-2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-5.12"), self.getLib().roundDown(self.makeNumber("-5.125"), self.makeNumber("2")))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.2E+2"), self.getLib().roundDown(self.makeNumber("125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("-1.2E+2"), self.getLib().roundDown(self.makeNumber("-125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().roundDown(self.makeNumber("125.125"), self.makeNumber("-2")))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().roundDown(self.makeNumber("-125.125"), self.makeNumber("-2")))

    def testRoundHalfUp(self):
        self.assertIsNone(self.getLib().roundHalfUp(None, None))
        self.assertIsNone(self.getLib().roundHalfUp(None, self.makeNumber("128")))
        self.assertIsNone(self.getLib().roundHalfUp(self.makeNumber("10"), None))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().roundHalfUp(self.makeNumber("5.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().roundHalfUp(self.makeNumber("2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().roundHalfUp(self.makeNumber("1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundHalfUp(self.makeNumber("1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundHalfUp(self.makeNumber("1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundHalfUp(self.makeNumber("-1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundHalfUp(self.makeNumber("-1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().roundHalfUp(self.makeNumber("-1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-3"), self.getLib().roundHalfUp(self.makeNumber("-2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-6"), self.getLib().roundHalfUp(self.makeNumber("-5.5"), self.makeNumber("0")))

        self.assertEqualsNumber(self.makeNumber("5.13"), self.getLib().roundHalfUp(self.makeNumber("5.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("2.13"), self.getLib().roundHalfUp(self.makeNumber("2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().roundHalfUp(self.makeNumber("1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundHalfUp(self.makeNumber("1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundHalfUp(self.makeNumber("1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundHalfUp(self.makeNumber("-1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundHalfUp(self.makeNumber("-1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().roundHalfUp(self.makeNumber("-1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-2.13"), self.getLib().roundHalfUp(self.makeNumber("-2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-5.13"), self.getLib().roundHalfUp(self.makeNumber("-5.125"), self.makeNumber("2")))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().roundHalfUp(self.makeNumber("125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().roundHalfUp(self.makeNumber("-125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().roundHalfUp(self.makeNumber("125.125"), self.makeNumber("-2")))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().roundHalfUp(self.makeNumber("-125.125"), self.makeNumber("-2")))

    def testRoundHalfDown(self):
        self.assertIsNone(self.getLib().roundHalfDown(None, None))
        self.assertIsNone(self.getLib().roundHalfDown(None, self.makeNumber("128")))
        self.assertIsNone(self.getLib().roundHalfDown(self.makeNumber("10"), None))

        # From RoundingMode.java https:#docs.oracle.com/javase/8/docs/api/java/math/RoundingMode.html
        self.assertEqualsNumber(self.makeNumber("5"), self.getLib().roundHalfDown(self.makeNumber("5.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().roundHalfDown(self.makeNumber("2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().roundHalfDown(self.makeNumber("1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundHalfDown(self.makeNumber("1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().roundHalfDown(self.makeNumber("1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundHalfDown(self.makeNumber("-1.0"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().roundHalfDown(self.makeNumber("-1.1"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().roundHalfDown(self.makeNumber("-1.6"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().roundHalfDown(self.makeNumber("-2.5"), self.makeNumber("0")))
        self.assertEqualsNumber(self.makeNumber("-5"), self.getLib().roundHalfDown(self.makeNumber("-5.5"), self.makeNumber("0")))

        self.assertEqualsNumber(self.makeNumber("5.12"), self.getLib().roundHalfDown(self.makeNumber("5.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("2.12"), self.getLib().roundHalfDown(self.makeNumber("2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.13"), self.getLib().roundHalfDown(self.makeNumber("1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundHalfDown(self.makeNumber("1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("1.12"), self.getLib().roundHalfDown(self.makeNumber("1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundHalfDown(self.makeNumber("-1.120"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.12"), self.getLib().roundHalfDown(self.makeNumber("-1.121"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-1.13"), self.getLib().roundHalfDown(self.makeNumber("-1.126"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-2.12"), self.getLib().roundHalfDown(self.makeNumber("-2.125"), self.makeNumber("2")))
        self.assertEqualsNumber(self.makeNumber("-5.12"), self.getLib().roundHalfDown(self.makeNumber("-5.125"), self.makeNumber("2")))

        # Negative scale
        self.assertEqualsNumber(self.makeNumber("1.3E+2"), self.getLib().roundHalfDown(self.makeNumber("125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("-1.3E+2"), self.getLib().roundHalfDown(self.makeNumber("-125.125"), self.makeNumber("-1")))
        self.assertEqualsNumber(self.makeNumber("1E+2"), self.getLib().roundHalfDown(self.makeNumber("125.125"), self.makeNumber("-2")))
        self.assertEqualsNumber(self.makeNumber("-1E+2"), self.getLib().roundHalfDown(self.makeNumber("-125.125"), self.makeNumber("-2")))

    def testFloor(self):
        self.assertIsNone(self.getLib().floor(None))
        self.assertIsNone(self.getLib().floor(None, None))

        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().floor(self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().floor(self.makeNumber(1.23)))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().floor(self.makeNumber(1.5)))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().floor(self.makeNumber(1.56)))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().floor(self.makeNumber(-1.56)))

        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().floor(self.makeNumber(1), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1.2"), self.getLib().floor(self.makeNumber(1.23), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1.5"), self.getLib().floor(self.makeNumber(1.5), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1.5"), self.getLib().floor(self.makeNumber(1.56), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("-1.6"), self.getLib().floor(self.makeNumber(-1.56), self.makeNumber(1)))

    def testCeiling(self):
        self.assertIsNone(self.getLib().ceiling(None))
        self.assertIsNone(self.getLib().ceiling(None, None))
        self.assertIsNone(self.getLib().ceiling(self.makeNumber(1), None))

        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().ceiling(self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().ceiling(self.makeNumber(1.23)))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().ceiling(self.makeNumber(1.5)))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().ceiling(self.makeNumber(1.56)))
        self.assertEqualsNumber(self.makeNumber("-1"), self.getLib().ceiling(self.makeNumber(-1.5)))

        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().ceiling(self.makeNumber(1), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1.3"), self.getLib().ceiling(self.makeNumber(1.23), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1.5"), self.getLib().ceiling(self.makeNumber(1.5), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("1.6"), self.getLib().ceiling(self.makeNumber(1.56), self.makeNumber(1)))
        self.assertEqualsNumber(self.makeNumber("-1.5"), self.getLib().ceiling(self.makeNumber(-1.56), self.makeNumber(1)))

    def testAbs(self):
        self.assertIsNone(self.getLib().abs(None))

        self.assertEqualsNumber(self.makeNumber("10"), self.getLib().abs(self.makeNumber(10)))
        self.assertEqualsNumber(self.makeNumber("10"), self.getLib().abs(self.makeNumber(-10)))

    def testModulo(self):
        self.assertIsNone(self.getLib().modulo(None, None))
        self.assertIsNone(self.getLib().modulo(self.makeNumber(10), self.makeNumber(0)))

        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().modulo(self.makeNumber(10), self.makeNumber(4)))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().modulo(self.makeNumber(10), self.makeNumber(-4)))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().modulo(self.makeNumber(-10), self.makeNumber(4)))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().modulo(self.makeNumber(-10), self.makeNumber(-4)))

        self.assertEqualsNumber(self.makeNumber("1.1"), self.getLib().modulo(self.makeNumber(10.1), self.makeNumber(4.5)))
        self.assertEqualsNumber(self.makeNumber("3.4"), self.getLib().modulo(self.makeNumber(-10.1), self.makeNumber(4.5)))
        self.assertEqualsNumber(self.makeNumber("-3.4"), self.getLib().modulo(self.makeNumber(10.1), self.makeNumber(-4.5)))
        self.assertEqualsNumber(self.makeNumber("-1.1"), self.getLib().modulo(self.makeNumber(-10.1), self.makeNumber(-4.5)))

    def testIntModulo(self):
        self.assertIsNone(self.getLib().modulo(None, None))
        self.assertIsNone(self.getLib().intModulo(self.makeNumber(10), self.makeNumber(0)))

        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().intModulo(self.makeNumber(10), self.makeNumber(4)))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().intModulo(self.makeNumber(10), self.makeNumber(-4)))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().intModulo(self.makeNumber(-10), self.makeNumber(4)))
        self.assertEqualsNumber(self.makeNumber("-2"), self.getLib().intModulo(self.makeNumber(-10), self.makeNumber(-4)))

    def testSqrt(self):
        self.assertIsNone(self.getLib().sqrt(None))
        self.assertIsNone(self.getLib().sqrt(self.makeNumber(-3)))

        self.assertEqualsNumber(self.makeNumber("4"), self.getLib().sqrt(self.makeNumber(16)))

    def testLog(self):
        self.assertIsNone(self.getLib().log(None))
        self.assertIsNone(self.getLib().log(self.makeNumber(-3)))

        self.assertEqualsNumber(self.makeNumber("2.30258509299404590109361379290930926799774169921875"), self.getLib().log(self.makeNumber(10)))

    def testExp(self):
        self.assertIsNone(self.getLib().exp(None))

        self.assertEqualsNumber(self.makeNumber("148.413159102576599934764089994132518768310546875"), self.getLib().exp(self.makeNumber(5)))

    def testOdd(self):
        self.assertIsNone(self.getLib().odd(None))

        self.assertFalse(self.getLib().odd(self.makeNumber("0.00")))
        self.assertTrue(self.getLib().odd(self.makeNumber("1.00")))
        self.assertTrue(self.getLib().odd(self.makeNumber(5)))
        self.assertFalse(self.getLib().odd(self.makeNumber(2)))

    def testEven(self):
        self.assertIsNone(self.getLib().even(None))

        self.assertTrue(self.getLib().even(self.makeNumber("0.00")))
        self.assertFalse(self.getLib().even(self.makeNumber("1.00")))
        self.assertFalse(self.getLib().even(self.makeNumber(5)))
        self.assertTrue(self.getLib().even(self.makeNumber(2)))

    #
    # String functions
    #
    def testContains(self):
        self.assertIsNone(self.getLib().contains(None, None))
        self.assertIsNone(self.getLib().contains(None, "bcg"))
        self.assertIsNone(self.getLib().contains("bcg", None))

        self.assertTrue(self.getLib().contains("abc", "a"))
        self.assertFalse(self.getLib().contains("aBc1", "bcg"))

        self.assertEqual(False, self.getLib().contains("foobar", "of"))
        self.assertEqual(True, self.getLib().contains("foobar", "fo"))
        self.assertEqual(True, self.getLib().contains("foobar", "r"))

    def testStartsWith(self):
        self.assertIsNone(self.getLib().startsWith(None, None))
        self.assertIsNone(self.getLib().startsWith(None, "bcg"))
        self.assertIsNone(self.getLib().startsWith("bcg", None))

        self.assertTrue(self.getLib().startsWith("abc", "a"))
        self.assertFalse(self.getLib().startsWith("aBc1", "bcg"))

        self.assertEqual(True, self.getLib().startsWith("foobar", "fo"))

    def testEndsWith(self):
        self.assertIsNone(self.getLib().endsWith(None, None))
        self.assertIsNone(self.getLib().endsWith(None, "bcg"))
        self.assertIsNone(self.getLib().endsWith("bcg", None))

        self.assertTrue(self.getLib().endsWith("abc", "c"))
        self.assertFalse(self.getLib().endsWith("aBc1", "bcg"))

        self.assertEqual(True, self.getLib().endsWith("foobar", "r"))

    def testSubstring(self):
        self.assertIsNone(self.getLib().substring(None, None))
        self.assertIsNone(self.getLib().substring(None, self.makeNumber("0")))
        self.assertIsNone(self.getLib().substring("", None))

        self.assertEqual("obar", self.getLib().substring("foobar", self.makeNumber("3")))
        self.assertEqual("oba", self.getLib().substring("foobar", self.makeNumber("3"), self.makeNumber("3")))
        self.assertEqual("a", self.getLib().substring("foobar", self.makeNumber("-2"), self.makeNumber("1")))

        # horse + grinning face emoji
        self.assertEqualsUnicodeString("\uD83D\uDE00", self.getLib().substring("foo\ud83d\udc0ebar\uD83D\uDE00", self.makeNumber("8")))
        self.assertEqualsUnicodeString("\uD83D\uDC0E", self.getLib().substring("foo\ud83d\udc0ebar\uD83D\uDE00", self.makeNumber("4"), self.makeNumber("1")))
        self.assertEqualsUnicodeString("\uD83D\uDC0Ebar", self.getLib().substring("foo\ud83d\udc0ebar\uD83D\uDE00", self.makeNumber("4"), self.makeNumber("4")))

    def testStringLength(self):
        self.assertIsNone(self.getLib().stringLength(None))

        self.assertEqualsNumber(self.makeNumber(3), self.getLib().stringLength("foo"))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\n"))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\r"))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\t"))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\""))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\'"))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\\"))
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\u0009"))
        self.assertEqualsNumber(self.makeNumber(6), self.getLib().stringLength("\\u0009"))
        # horse emoji
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\uD83D\uDCA9"))
        # horse emoji lowercase
        self.assertEqualsNumber(self.makeNumber(1), self.getLib().stringLength("\ud83d\udca9"))
        # horse + grinning face emoji
        self.assertEqualsNumber(self.makeNumber(2), self.getLib().stringLength("\ud83d\udc0e\uD83D\uDE00"))
        # horse + grinning face emoji
        self.assertEqualsNumber(self.makeNumber(2), self.getLib().stringLength("🐎😀"))

    def testUpperCase(self):
        self.assertIsNone(self.getLib().upperCase(None))

        self.assertEqual("ABC4", self.getLib().upperCase("aBc4"))

    def testLowerCase(self):
        self.assertIsNone(self.getLib().lowerCase(None))

        self.assertEqual("abc4", self.getLib().lowerCase("aBc4"))

    def testSubstringBefore(self):
        self.assertIsNone(self.getLib().substringBefore(None, None))
        self.assertIsNone(self.getLib().substringBefore(None, "bar"))
        self.assertIsNone(self.getLib().substringBefore("foobar", None))

        self.assertEqual("foo", self.getLib().substringBefore("foobar", "bar"))
        self.assertEqual("", self.getLib().substringBefore("foobar", "xyz"))

    def testSubstringAfter(self):
        self.assertIsNone(self.getLib().substringAfter(None, None))
        self.assertIsNone(self.getLib().substringAfter(None, "ob"))
        self.assertIsNone(self.getLib().substringAfter("foobar", None))

        self.assertEqual("ar", self.getLib().substringAfter("foobar", "ob"))
        self.assertEqual("", self.getLib().substringAfter("foobar", "xyz"))

    def testReplace(self):
        self.assertIsNone(self.getLib().replace("", "", ""))
        self.assertIsNone(self.getLib().replace("", "", None))

        self.assertIsNone(self.getLib().replace("abc", "[a-z]*", "#"))
        self.assertEqual("#", self.getLib().replace("abc", "[a-z]+", "#"))

        self.assertIsNone(self.getLib().replace(None, "(ab)|(a)", "[1=$1][2=$2]"))
        self.assertIsNone(self.getLib().replace("abcd", None, "[1=$1][2=$2]"))
        self.assertIsNone(self.getLib().replace("abcd", "(ab)|(a)", None))

        self.assertEqual("[1=ab][2=]cd", self.getLib().replace("abcd", "(ab)|(a)", "[1=$1][2=$2]"))

        self.assertEqual("a*cada*", self.getLib().replace("abracadabra", "bra", "*"))
        self.assertEqual("*", self.getLib().replace("abracadabra", "a.*a", "*"))
        self.assertEqual("*c*bra", self.getLib().replace("abracadabra", "a.*?a", "*"))
        self.assertEqual("brcdbr", self.getLib().replace("abracadabra", "a", ""))
        self.assertEqual("abbraccaddabbra", self.getLib().replace("abracadabra", "a(.)", "a$1$1"))
        self.assertEqual(None, self.getLib().replace("abracadabra", ".*?", "$1"))
        self.assertEqual("b", self.getLib().replace("AAAA", "A+", "b"))
        self.assertEqual("bbbb", self.getLib().replace("AAAA", "A+?", "b"))
        self.assertEqual("carted", self.getLib().replace("darted", "^(.*?)d(.*)$", "$1c$2"))

    def testMatches(self):
        self.assertTrue(self.getLib().matches("", "", ""))
        self.assertTrue(self.getLib().matches("", "", None))

        self.assertTrue(self.getLib().matches("", "[a-z]*", ""))
        self.assertTrue(self.getLib().matches("abc", "[a-z]+", ""))

        self.assertTrue(self.getLib().matches("abracadabra", "bra"))
        self.assertTrue(self.getLib().matches("abracadabra", "^a.*a$"))
        self.assertFalse(self.getLib().matches("abracadabra", "^bra"))

        input: STRING = "Kaum hat dies der Hahn gesehen,\n" \
                        "Fangt er auch schon an zu krahen:\n" \
                        "Kikeriki! Kikikerikih!!\n" \
                        "Tak, tak, tak! - da kommen sie.\n"

        self.assertFalse(self.getLib().matches(input, "Kaum.*krahen"))
        self.assertTrue(self.getLib().matches(input, "Kaum.*krahen", "s"))
        self.assertTrue(self.getLib().matches(input, "^Kaum.*gesehen,$", "m"))
        self.assertFalse(self.getLib().matches(input, "^Kaum.*gesehen,$"))
        self.assertTrue(self.getLib().matches(input, "kiki", "i"))

    def testSplit(self):
        self.assertIsNone(self.getLib().split(None, None))
        self.assertIsNone(self.getLib().split("", ""))

        self.assertEqual(self.getLib().asList("John", "Doe"), self.getLib().split("John Doe", "\\s"))
        self.assertEqual(self.getLib().asList("a", "b", "c", "", ""), self.getLib().split("a;b;c;;", ";"))

    #
    # Boolean functions
    #
    def testNot(self):
        self.assertTrue(self.getLib().not_(False))
        self.assertFalse(self.getLib().not_(True))
        self.assertIsNone(self.getLib().not_(None))

    #
    # Date properties
    #
    def testDateProperties(self):
        self.assertIsNone(self.getLib().year(None))
        self.assertEqualsNumber(self.makeNumber("2018"), self.getLib().year(self.getLib().date("2018-12-10")))

        self.assertIsNone(self.getLib().month(None))
        self.assertEqualsNumber(self.makeNumber("12"), self.getLib().month(self.getLib().date("2018-12-10")))

        self.assertIsNone(self.getLib().day(None))
        self.assertEqualsNumber(self.makeNumber("10"), self.getLib().day(self.getLib().date("2018-12-10")))

        self.assertIsNone(self.getLib().weekday(None))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().weekday(self.getLib().date("2018-12-10")))

    #
    # Time properties
    #
    def testTimeProperties(self):
        # See each implementation
        self.assertTrue(True)

    #
    # Date and time properties
    #
    def testDateAndTimeProperties(self):
        # See each implementation
        self.assertTrue(True)

    #
    # Duration properties
    #
    def testDurationProperties(self):
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().years(self.makeDuration("P1Y2M")))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().months(self.makeDuration("P1Y2M")))

        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().days(self.makeDuration("P3DT4H5M6.700S")))
        self.assertEqualsNumber(self.makeNumber("4"), self.getLib().hours(self.makeDuration("P3DT4H5M6.700S")))
        self.assertEqualsNumber(self.makeNumber("5"), self.getLib().minutes(self.makeDuration("P3DT4H5M6.700S")))
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().seconds(self.makeDuration("P3DT4H5M6.700S")))

    #
    # Date and time functions
    #
    def testDateAndTimeFunctions(self):
        self.assertTrue(self.getLib().is_(None, None))
        self.assertFalse(self.getLib().is_(None, self.makeDate("2010-10-10")))
        self.assertFalse(self.getLib().is_(self.makeTime("12:00:00"), None))
        self.assertFalse(self.getLib().is_(self.makeDate("2012-12-25"), self.makeTime("23:00:50")))

        self.assertTrue(self.getLib().is_(self.makeNumber("10"), self.makeNumber("10")))
        self.assertFalse(self.getLib().is_(self.makeNumber("10"), self.makeNumber("11")))

        self.assertTrue(self.getLib().is_(True, True))
        self.assertFalse(self.getLib().is_(True, False))

        self.assertTrue(self.getLib().is_("abc", "abc"))
        self.assertFalse(self.getLib().is_("abc", "abd"))

        self.assertTrue(self.getLib().is_(self.makeDate("2012-12-25"), self.makeDate("2012-12-25")))
        self.assertFalse(self.getLib().is_(self.makeDate("2012-12-25"), self.makeDate("2012-12-26")))

        self.assertTrue(self.getLib().is_(self.makeTime("23:00:50z"), self.makeTime("23:00:50z")))
        self.assertTrue(self.getLib().is_(self.makeTime("23:00:50z"), self.makeTime("23:00:50Z")))
#        self.assertTrue(self.getLib().is_(self.makeTime("23:00:50z"), self.makeTime("23:00:50+00:00")))

        self.assertTrue(self.getLib().is_(self.makeDateAndTime("2012-12-25T12:00:00"), self.makeDateAndTime("2012-12-25T12:00:00")))
        self.assertFalse(self.getLib().is_(self.makeDateAndTime("2012-12-25T12:00:00"), self.makeDateAndTime("2012-12-26T12:00:00z")))

        self.assertTrue(self.getLib().is_(self.makeDuration("P1Y"), self.makeDuration("P1Y")))
        self.assertFalse(self.getLib().is_(self.makeDuration("P1Y"), self.makeDuration("-P1Y")))
        self.assertTrue(self.getLib().is_(self.makeDuration("P1Y"), self.makeDuration("P12M")))

        self.assertTrue(self.getLib().is_([], []))
        self.assertTrue(self.getLib().is_([self.makeNumber(1), self.makeNumber(3)], [self.makeNumber(1), self.makeNumber(3)]))
        self.assertFalse(self.getLib().is_([self.makeNumber(1), self.makeNumber(3)], []))

        self.assertTrue(self.getLib().is_(Range(True, self.makeNumber(1), True, self.makeNumber(3)), Range(True, self.makeNumber(1), True, self.makeNumber(3))))
        self.assertFalse(self.getLib().is_(Range(True, self.makeNumber(1), True, self.makeNumber(3)), Range(True, self.makeNumber(2), True, self.makeNumber(3))))
        self.assertFalse(self.getLib().is_(Range(True, self.makeNumber(1), True, self.makeNumber(3)), Range(False, self.makeNumber(1), True, self.makeNumber(3))))

        self.assertTrue(self.getLib().is_(Context(), Context()))
        self.assertTrue(self.getLib().is_(Context().add("a", self.makeNumber(1)), Context().add("a", self.makeNumber(1))))
        self.assertFalse(self.getLib().is_(Context().add("a", self.makeNumber(1)), Context()))

    #
    # Temporal functions
    #
    def testTemporalFunctions(self):
        self.assertIsNone(self.getLib().dayOfYear(None))
        self.assertIsNone(self.getLib().dayOfWeek(None))
        self.assertIsNone(self.getLib().weekOfYear(None))
        self.assertIsNone(self.getLib().monthOfYear(None))

        self.assertEqual(self.makeNumber("260"), self.getLib().dayOfYear(self.makeDate("2019-09-17")))
        self.assertEqual("Tuesday", self.getLib().dayOfWeek(self.makeDate("2019-09-17")))
        self.assertEqual(self.makeNumber("38"), self.getLib().weekOfYear(self.makeDate("2019-09-17")))
        self.assertEqual(self.makeNumber("1"), self.getLib().weekOfYear(self.makeDate("2003-12-29")))
        self.assertEqual(self.makeNumber("1"), self.getLib().weekOfYear(self.makeDate("2004-01-04")))
        self.assertEqual(self.makeNumber("53"), self.getLib().weekOfYear(self.makeDate("2005-01-01")))
        self.assertEqual(self.makeNumber("1"), self.getLib().weekOfYear(self.makeDate("2005-01-03")))
        self.assertEqual(self.makeNumber("1"), self.getLib().weekOfYear(self.makeDate("2005-01-09")))
        self.assertEqual("September", self.getLib().monthOfYear(self.makeDate("2019-09-17")))

    #
    # List functions
    #
    def testListContains(self):
        self.assertIsNone(self.getLib().listContains(None, self.makeNumber(2)))
        self.assertIsNone(self.getLib().listContains(None, self.makeNumber(2)))

        self.assertEqual(False, self.getLib().listContains(self.makeNumberList(1, 2, 3), None))
        self.assertEqual(True, self.getLib().listContains(self.makeNumberList(1, 2, 3), self.makeNumber(2)))

    def testAppend(self):
        self.assertEqualsList("[None]", self.getLib().append(None, None))
        self.assertEqual(["3"], self.getLib().append(None, "3"))
        self.assertEqual(self.makeNumberList("1", None, "3", None, "3"), self.getLib().append(self.makeNumberList(1, None, 3), None, self.makeNumber(3)))

        self.assertEqual(self.makeNumberList("1", "2", "3", "4"), self.getLib().append(self.makeNumberList(1, 2, 3), self.makeNumber(4)))
        self.assertEqual(self.makeNumberList("1", None, "3", "4"), self.getLib().append(self.makeNumberList(1, None, 3), self.makeNumber(4)))

    def testMin(self):
        super().testMin()
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().min(self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)))
        self.assertIsNone(self.getLib().min(1, None, 3))

    def testMax(self):
        super().testMax()
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().max(self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)))
        self.assertIsNone(self.getLib().max(1, None, 3))

    def testSum(self):
        super().testSum()
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().sum(self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)))
        self.assertIsNone(self.getLib().sum(1, None, 3))

    def testMean(self):
        self.assertIsNone(self.getLib().mean())
        self.assertIsNone(self.getLib().mean(None))
        self.assertIsNone(self.getLib().mean([]))
        self.assertIsNone(self.getLib().mean(self.makeNumberList()))

        self.assertIsNone(self.getLib().mean(self.makeNumber(1), None, self.makeNumber(3)))
        self.assertIsNone(self.getLib().mean(self.makeNumberList(1, None, 3)))

        self.assertEqualsNumber(2.0, self.getLib().mean(self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)), 0.001)
        self.assertEqualsNumber(2.0, self.getLib().mean(self.makeNumberList(1, 2, 3)), 0.001)

    def testAll(self):
        self.assertFalse(self.getLib().all(True, True, False))
        self.assertFalse(self.getLib().all(None, False))
        self.assertFalse(self.getLib().all([True, True, False]))
        self.assertFalse(self.getLib().all([None, False]))

        self.assertTrue(self.getLib().all())
        self.assertTrue(self.getLib().all([]))
        self.assertTrue(self.getLib().all(True, True))
        self.assertTrue(self.getLib().all([True, True]))

        self.assertIsNone(self.getLib().all(None))
        self.assertIsNone(self.getLib().all(None, None))
        self.assertIsNone(self.getLib().all(None, True))
        self.assertIsNone(self.getLib().all([None, None]))
        self.assertIsNone(self.getLib().all([None, True]))

    def testAny(self):
        self.assertTrue(self.getLib().any(True, True))
        self.assertTrue(self.getLib().any(False, True))
        self.assertTrue(self.getLib().any(None, True))
        self.assertTrue(self.getLib().any([True, True]))
        self.assertTrue(self.getLib().any([False, True]))
        self.assertTrue(self.getLib().any([None, True]))

        self.assertFalse(self.getLib().any())
        self.assertFalse(self.getLib().any([]))
        self.assertFalse(self.getLib().any(False, False, False))
        self.assertFalse(self.getLib().any([False, False, False]))

        self.assertIsNone(self.getLib().any(None))
        self.assertIsNone(self.getLib().any(None, None))
        self.assertIsNone(self.getLib().any(None, False))
        self.assertIsNone(self.getLib().any([None, None]))
        self.assertIsNone(self.getLib().any([None, False]))

    def testSublist(self):
        self.assertIsNone(self.getLib().sublist(None, None))
        self.assertIsNone(self.getLib().sublist(None, None, None))

        self.assertEqual(self.makeNumberList("1", "2", "3"), self.getLib().sublist(self.makeNumberList(1, 2, 3), self.makeNumber("1")))
        self.assertEqual(self.makeNumberList("1", "2"), self.getLib().sublist(self.makeNumberList(1, 2, 3), self.makeNumber("1"), self.makeNumber("2")))
        self.assertEqual(self.makeNumberList("2"), self.getLib().sublist(self.makeNumberList(1, 2, 3), self.makeNumber("2"), self.makeNumber("1")))

    def testConcatenate(self):
        self.assertIsNone(self.getLib().concatenate(None, None))

        self.assertEqual(self.makeNumberList("1", "2", "3", "4", "5", "6"), self.getLib().concatenate(self.makeNumberList(1, 2, 3), self.makeNumberList(4, 5, 6)))

    def testInsertBefore(self):
        self.assertIsNone(self.getLib().insertBefore(None, None, None))

        self.assertEqual(self.makeNumberList("2", "1", "3"), self.getLib().insertBefore(self.makeNumberList(1, 3), self.makeNumber("1"), self.makeNumber("2")))
        self.assertEqual(self.makeNumberList("1", "2", "4", "3"), self.getLib().insertBefore(self.makeNumberList(1, 2, 3), self.makeNumber("-1"), self.makeNumber("4")))
        # Out of bounds
        self.assertEqual(self.makeNumberList("1", "2", "3"), self.getLib().insertBefore(self.makeNumberList(1, 2, 3), self.makeNumber("5"), self.makeNumber("4")))
        self.assertEqual(self.makeNumberList("1", "2", "3"), self.getLib().insertBefore(self.makeNumberList(1, 2, 3), self.makeNumber("-5"), self.makeNumber("4")))

    def testRemove(self):
        self.assertIsNone(self.getLib().remove(None, None))

        self.assertEqual(self.makeNumberList("1", "3"), self.getLib().remove(self.makeNumberList(1, 2, 3), self.makeNumber("2")))

    def testReverse(self):
        self.assertEqual([], self.getLib().reverse(None))

        self.assertEqual(self.makeNumberList("3", "2", "1"), self.getLib().reverse(self.makeNumberList(1, 2, 3)))

    def testIndexOf(self):
        self.assertEqual([], self.getLib().indexOf(None, None))

        self.assertEqual(self.makeNumberList("2", "4"), self.getLib().indexOf(["1", "2", "3", "2"], "2"))

    def testUnion(self):
        self.assertIsNone(self.getLib().union(None, None))

        self.assertEqual(self.makeNumberList("1", "2", "3"), self.getLib().union(self.makeNumberList(1, 2), self.makeNumberList(2, 3)))

    def testDistinctValues(self):
        self.assertEqual([], self.getLib().distinctValues(None))

        self.assertEqual(self.makeNumberList("1", "2", "3"), self.getLib().distinctValues(self.makeNumberList(1, 2, 3, 2, 1)))

    def testFlatten(self):
        self.assertIsNone(self.getLib().flatten(None))

        self.assertEqual(self.makeList("1", "2", "3", "4"), self.getLib().flatten([
            [["1", "2"]],
            [["3"]],
            "4"
        ]))

    def testProduct(self):
        self.assertIsNone(self.getLib().product())
        self.assertIsNone(self.getLib().product(None))
        self.assertIsNone(self.getLib().product([]))
        self.assertIsNone(self.getLib().product((self.makeNumberList())))

        self.assertIsNone(self.getLib().product(self.makeNumber(2), None, self.makeNumber(4)))
        self.assertIsNone(self.getLib().product(self.makeNumberList(2, None, 4)))

        self.assertEqualsNumber(self.makeNumber(24), self.getLib().product(self.makeNumber(2), self.makeNumber(3), self.makeNumber(4)))
        self.assertEqualsNumber(self.makeNumber(24), self.getLib().product(self.makeNumberList(2, 3, 4)))

    def testMedian(self):
        self.assertIsNone(self.getLib().median())
        self.assertIsNone(self.getLib().median(None))
        self.assertIsNone(self.getLib().median([]))
        self.assertIsNone(self.getLib().median(self.makeNumberList()))

        self.assertIsNone(self.getLib().median(self.makeNumber(8), None, self.makeNumber(5)))
        self.assertIsNone(self.getLib().median(self.makeNumberList(8, None, 5)))

        self.assertEqualsNumber(self.makeNumber(4), self.getLib().median(self.makeNumber(8), self.makeNumber(2), self.makeNumber(5), self.makeNumber(3), self.makeNumber(4)))
        self.assertEqualsNumber(self.makeNumber(4), self.getLib().median(self.makeNumberList(8, 2, 5, 3, 4)))
        self.assertEqualsNumber(self.makeNumber(2.5), self.getLib().median(self.makeNumber(6), self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)))
        self.assertEqualsNumber(self.makeNumber(2.5), self.getLib().median(self.makeNumberList(6, 1, 2, 3)))

    def testStddev(self):
        self.assertIsNone(self.getLib().stddev())
        self.assertIsNone(self.getLib().stddev(None))
        self.assertIsNone(self.getLib().stddev([]))
        self.assertIsNone(self.getLib().stddev(self.makeNumberList()))

        self.assertIsNone(self.getLib().stddev(self.makeNumber(2), self.makeNumber(4), None, self.makeNumber(5)))
        self.assertIsNone(self.getLib().stddev(self.makeNumberList(2, 4, None, 5)))

        self.assertEqualsNumber(self.makeNumber("2.0816659994661"), self.getLib().stddev(self.makeNumber(2), self.makeNumber(4), self.makeNumber(7), self.makeNumber(5)))
        self.assertEqualsNumber(self.makeNumber("2.0816659994661"), self.getLib().stddev(self.makeNumberList(2, 4, 7, 5)))

    def testMode(self):
        self.assertIsNone(self.getLib().mode(None))

        self.assertEqual(self.makeNumberList(), self.getLib().mode())
        self.assertEqual(self.makeNumberList(), self.getLib().mode([]))
        self.assertEqual(self.makeNumberList(), self.getLib().mode(self.makeNumberList()))

        self.assertIsNone(self.getLib().mode(self.makeNumber(1), None))
        self.assertIsNone(self.getLib().mode(self.makeNumberList(1, None)))

        self.assertEqual(self.makeNumberList(6), self.getLib().mode(self.makeNumber(6), self.makeNumber(3), self.makeNumber(9), self.makeNumber(6), self.makeNumber(6)))
        self.assertEqual(self.makeNumberList(6), self.getLib().mode(self.makeNumberList(6, 3, 9, 6, 6)))
        self.assertEqual(self.makeNumberList(1, 6), self.getLib().mode(self.makeNumber(6), self.makeNumber(1), self.makeNumber(9), self.makeNumber(6), self.makeNumber(1)))
        self.assertEqual(self.makeNumberList(1, 6), self.getLib().mode(self.makeNumberList(6, 1, 9, 6, 1)))

    def testCollect(self):
        self.getLib().collect(None, None)
        self.getLib().collect([], None)

        result = []
        self.getLib().collect(result, [["1", "2"], "3"])
        self.assertEqual(["1", "2", "3"], result)

    def testSort(self):
        comparator = LambdaExpression(lambda *args: (x := args[0], y := args[1], self.getLib().stringLessThan(x, y)))

        self.assertIsNone(self.getLib().sort(None, None))
        self.assertIsNone(self.getLib().sort(None, comparator))
        self.assertEqual([], self.getLib().sort([], None))

        self.assertEqual(["1", "2", "3"], self.getLib().sort(["3", "1", "2"], comparator))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from unittest import TestCase

from jdmn.feel.lib.DefaultStandardFEELLib import DefaultStandardFEELLib
from jdmn.feel.lib.Types import RANGE, STRING, COMPARABLE
from jdmn.runtime.Range import Range


class AbstractRangeLibTest(TestCase):
    """
    Base test class for RangeLib
    """
    __test__ = False

    feelLib = DefaultStandardFEELLib()

    def testBefore(self):
        self.assertIsNone(self.feelLib.before(None, None))

        self.assertTrue(self.feelLib.before(self.makePoint(1), self.makePoint(10)))
        self.assertFalse(self.feelLib.before(self.makePoint(10), self.makePoint(1)))
        self.assertFalse(self.feelLib.before(self.makePoint(1), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.before(self.makePoint(1), self.makeRange("(", 1, 10, "]")))
        self.assertTrue(self.feelLib.before(self.makePoint(1), self.makeRange("[", 5, 10, "]")))
        self.assertFalse(self.feelLib.before(self.makeRange("[", 1, 10, "]"), self.makePoint(10)))
        self.assertTrue(self.feelLib.before(self.makeRange("[", 1, 10, ")"), self.makePoint(10)))
        self.assertTrue(self.feelLib.before(self.makeRange("[", 1, 10, "]"), self.makePoint(15)))
        self.assertTrue(self.feelLib.before(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 15, 20, "]")))
        self.assertFalse(self.feelLib.before(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 10, 20, "]")))
        self.assertTrue(self.feelLib.before(self.makeRange("[", 1, 10, ")"), self.makeRange("[", 10, 20, "]")))
        self.assertTrue(self.feelLib.before(self.makeRange("[", 1, 10, "]"), self.makeRange("(", 10, 20, "]")))

    def testAfter(self):
        self.assertIsNone(self.feelLib.after(None, None))
        self.assertIsNone(self.feelLib.after(None, None))
        self.assertIsNone(self.feelLib.after(None, None))
        self.assertIsNone(self.feelLib.after(None, None))

        self.assertTrue(self.feelLib.after(self.makePoint(10), self.makePoint(5)))
        self.assertFalse(self.feelLib.after(self.makePoint(5), self.makePoint(10)))
        self.assertTrue(self.feelLib.after(self.makePoint(12), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.after(self.makePoint(10), self.makeRange("[", 1, 10, ")")))
        self.assertFalse(self.feelLib.after(self.makePoint(10), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.after(self.makeRange("[", 11, 20, "]"), self.makePoint(12)))
        self.assertTrue(self.feelLib.after(self.makeRange("[", 11, 20, "]"), self.makePoint(10)))
        self.assertTrue(self.feelLib.after(self.makeRange("(", 11, 20, "]"), self.makePoint(11)))
        self.assertFalse(self.feelLib.after(self.makeRange("[", 11, 20, "]"), self.makePoint(11)))
        self.assertTrue(self.feelLib.after(self.makeRange("[", 11, 20, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.after(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 11, 20, "]")))
        self.assertTrue(self.feelLib.after(self.makeRange("[", 11, 20, "]"), self.makeRange("[", 1, 11, ")")))
        self.assertTrue(self.feelLib.after(self.makeRange("(", 11, 20, "]"), self.makeRange("[", 1, 11, "]")))

    def testMeets(self):
        self.assertIsNone(self.feelLib.meets(None, None))

        self.assertTrue(self.feelLib.meets(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 5, 10, "]")))
        self.assertFalse(self.feelLib.meets(self.makeRange("[", 1, 5, ")"), self.makeRange("[", 5, 10, "]")))
        self.assertFalse(self.feelLib.meets(self.makeRange("[", 1, 5, "]"), self.makeRange("(", 5, 10, "]")))
        self.assertFalse(self.feelLib.meets(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 6, 10, "]")))

    def testMetBy(self):
        self.assertIsNone(self.feelLib.metBy(None, None))

        self.assertTrue(self.feelLib.metBy(self.makeRange("[", 5, 10, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.metBy(self.makeRange("[", 5, 10, "]"), self.makeRange("[", 1, 5, ")")))
        self.assertFalse(self.feelLib.metBy(self.makeRange("(", 5, 10, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.metBy(self.makeRange("[", 6, 10, "]"), self.makeRange("[", 1, 5, "]")))

    def testOverlaps(self):
        self.assertIsNone(self.feelLib.overlaps(None, None))

        self.assertTrue(self.feelLib.overlaps(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 3, 8, "]")))
        self.assertTrue(self.feelLib.overlaps(self.makeRange("[", 3, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertTrue(self.feelLib.overlaps(self.makeRange("[", 1, 8, "]"), self.makeRange("[", 3, 5, "]")))
        self.assertTrue(self.feelLib.overlaps(self.makeRange("[", 3, 5, "]"), self.makeRange("[", 1, 8, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 6, 8, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("[", 6, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertTrue(self.feelLib.overlaps(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 5, 8, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("[", 1, 5, "]"), self.makeRange("(", 5, 8, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("[", 1, 5, ")"), self.makeRange("[", 5, 8, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("[", 1, 5, ")"), self.makeRange("(", 5, 8, "]")))
        self.assertTrue(self.feelLib.overlaps(self.makeRange("[", 5, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("(", 5, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("[", 5, 8, "]"), self.makeRange("[", 1, 5, ")")))
        self.assertFalse(self.feelLib.overlaps(self.makeRange("(", 5, 8, "]"), self.makeRange("[", 1, 5, ")")))

    def testOverlapsBefore(self):
        self.assertIsNone(self.feelLib.overlapsBefore(None, None))

        self.assertTrue(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 3, 8, "]")))
        self.assertFalse(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 6, 8, "]")))
        self.assertTrue(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 5, 8, "]")))
        self.assertFalse(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, "]"), self.makeRange("(", 5, 8, "]")))
        self.assertFalse(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, ")"), self.makeRange("[", 5, 8, "]")))
        self.assertTrue(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, ")"), self.makeRange("(", 1, 5, "]")))
        self.assertTrue(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, "]"), self.makeRange("(", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, ")"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlapsBefore(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 1, 5, "]")))

    def testOverlapsAfter(self):
        self.assertIsNone(self.feelLib.overlapsAfter(None, None))

        self.assertTrue(self.feelLib.overlapsAfter(self.makeRange("[", 3, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlapsAfter(self.makeRange("[", 6, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertTrue(self.feelLib.overlapsAfter(self.makeRange("[", 5, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlapsAfter(self.makeRange("(", 5, 8, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlapsAfter(self.makeRange("[", 5, 8, "]"), self.makeRange("[", 1, 5, ")")))
        self.assertTrue(self.feelLib.overlapsAfter(self.makeRange("(", 1, 5, "]"), self.makeRange("[", 1, 5, ")")))
        self.assertTrue(self.feelLib.overlapsAfter(self.makeRange("(", 1, 5, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.overlapsAfter(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 1, 5, ")")))
        self.assertFalse(self.feelLib.overlapsAfter(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 1, 5, "]")))

    def testFinishes(self):
        self.assertIsNone(self.feelLib.finishes(None, None))
        self.assertIsNone(self.feelLib.finishes(None, None))

        self.assertTrue(self.feelLib.finishes(self.makePoint(10), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.finishes(self.makePoint(10), self.makeRange("[", 1, 10, ")")))
        self.assertTrue(self.feelLib.finishes(self.makeRange("[", 5, 10, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.finishes(self.makeRange("[", 5, 10, ")"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.finishes(self.makeRange("[", 5, 10, ")"), self.makeRange("[", 1, 10, ")")))
        self.assertTrue(self.feelLib.finishes(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.finishes(self.makeRange("(", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))

    def testFinishedBy(self):
        self.assertIsNone(self.feelLib.finishedBy(None, None))
        self.assertIsNone(self.feelLib.finishedBy(None, None))

        self.assertTrue(self.feelLib.finishedBy(self.makeRange("[", 1, 10, "]"), self.makePoint(10)))
        self.assertFalse(self.feelLib.finishedBy(self.makeRange("[", 1, 10, ")"), self.makePoint(10)))
        self.assertTrue(self.feelLib.finishedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 5, 10, "]")))
        self.assertFalse(self.feelLib.finishedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 5, 10, ")")))
        self.assertTrue(self.feelLib.finishedBy(self.makeRange("[", 1, 10, ")"), self.makeRange("[", 5, 10, ")")))
        self.assertTrue(self.feelLib.finishedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.finishedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("(", 1, 10, "]")))

    def testIncludes(self):
        self.assertIsNone(self.feelLib.includes(None, None))
        self.assertIsNone(self.feelLib.includes(None, None))

        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makePoint(5)))
        self.assertFalse(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makePoint(12)))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makePoint(1)))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makePoint(10)))
        self.assertFalse(self.feelLib.includes(self.makeRange("(", 1, 10, "]"), self.makePoint(1)))
        self.assertFalse(self.feelLib.includes(self.makeRange("[", 1, 10, ")"), self.makePoint(10)))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 4, 6, "]")))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertTrue(self.feelLib.includes(self.makeRange("(", 1, 10, "]"), self.makeRange("(", 1, 5, "]")))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makeRange("(", 1, 10, ")")))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, ")"), self.makeRange("[", 5, 10, ")")))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, ")")))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makeRange("(", 1, 10, "]")))
        self.assertTrue(self.feelLib.includes(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))

    def testDuring(self):
        self.assertIsNone(self.feelLib.during(None, None))
        self.assertIsNone(self.feelLib.during(None, None))

        self.assertTrue(self.feelLib.during(self.makePoint(5), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.during(self.makePoint(12), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makePoint(1), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makePoint(10), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.during(self.makePoint(1), self.makeRange("(", 1, 10, "]")))
        self.assertFalse(self.feelLib.during(self.makePoint(10), self.makeRange("[", 1, 10, ")")))
        self.assertTrue(self.feelLib.during(self.makeRange("[", 4, 6, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makeRange("(", 1, 5, "]"), self.makeRange("(", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makeRange("(", 1, 10, ")"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makeRange("[", 5, 10, ")"), self.makeRange("[", 1, 10, ")")))
        self.assertTrue(self.feelLib.during(self.makeRange("[", 1, 10, ")"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makeRange("(", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.during(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))

    def testStarts(self):
        self.assertIsNone(self.feelLib.starts(None, None))
        self.assertIsNone(self.feelLib.starts(None, None))

        self.assertTrue(self.feelLib.starts(self.makePoint(1), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.starts(self.makePoint(1), self.makeRange("(", 1, 10, "]")))
        self.assertFalse(self.feelLib.starts(self.makePoint(2), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.starts(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.starts(self.makeRange("(", 1, 5, "]"), self.makeRange("(", 1, 10, "]")))
        self.assertFalse(self.feelLib.starts(self.makeRange("(", 1, 5, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertFalse(self.feelLib.starts(self.makeRange("[", 1, 5, "]"), self.makeRange("(", 1, 10, "]")))
        self.assertTrue(self.feelLib.starts(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.starts(self.makeRange("[", 1, 10, ")"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.starts(self.makeRange("(", 1, 10, ")"), self.makeRange("(", 1, 10, ")")))

    def testStartedBy(self):
        self.assertIsNone(self.feelLib.startedBy(None, None))
        self.assertIsNone(self.feelLib.startedBy(None, None))

        self.assertTrue(self.feelLib.startedBy(self.makeRange("[", 1, 10, "]"), self.makePoint(1)))
        self.assertFalse(self.feelLib.startedBy(self.makeRange("(", 1, 10, "]"), self.makePoint(1)))
        self.assertFalse(self.feelLib.startedBy(self.makeRange("[", 1, 10, "]"), self.makePoint(2)))
        self.assertTrue(self.feelLib.startedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertTrue(self.feelLib.startedBy(self.makeRange("(", 1, 10, "]"), self.makeRange("(", 1, 5, "]")))
        self.assertFalse(self.feelLib.startedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("(", 1, 5, "]")))
        self.assertFalse(self.feelLib.startedBy(self.makeRange("(", 1, 10, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertTrue(self.feelLib.startedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, "]")))
        self.assertTrue(self.feelLib.startedBy(self.makeRange("[", 1, 10, "]"), self.makeRange("[", 1, 10, ")")))
        self.assertTrue(self.feelLib.startedBy(self.makeRange("(", 1, 10, ")"), self.makeRange("(", 1, 10, ")")))

    def testCoincides(self):
        self.assertIsNone(self.feelLib.coincides(None, None))
        self.assertIsNone(self.feelLib.coincides(None, None))

        self.assertTrue(self.feelLib.coincides(self.makePoint(5), self.makePoint(5)))
        self.assertFalse(self.feelLib.coincides(self.makePoint(3), self.makePoint(4)))
        self.assertTrue(self.feelLib.coincides(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.coincides(self.makeRange("(", 1, 5, ")"), self.makeRange("[", 1, 5, "]")))
        self.assertFalse(self.feelLib.coincides(self.makeRange("[", 1, 5, "]"), self.makeRange("[", 2, 6, "]")))

    def makeRange(self, startIncluded: STRING, start: int, end: int, endIncluded: STRING) -> RANGE:
        return Range(startIncluded == "[", self.makePoint(start), endIncluded == "]", self.makePoint(end))

    def makePoint(self, number: int) -> COMPARABLE:
        pass



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from collections import defaultdict
from decimal import Decimal
from typing import List, Any, Optional

from jdmn.feel.lib.Types import STRING, NUMBER, BOOLEAN
from jdmn.feel.lib.Utils import varArgToList
from jdmn.feel.lib.type.numeric.DefaultNumericType import DefaultNumericType
from jdmn.runtime.NumericRoundingMode import NumericRoundingMode


class DefaultNumericLib:
    def number(self, literal: STRING, groupingSeparator: STRING = None, decimalSeparator: STRING = None) -> NUMBER:
        if not literal:
            return None
        if not (" " == groupingSeparator or "." == groupingSeparator or "," == groupingSeparator or groupingSeparator is None):
            return None
        if not ("." == decimalSeparator or "," == decimalSeparator or decimalSeparator is None):
            return None
        if groupingSeparator is not None and groupingSeparator == decimalSeparator:
            return None

        if groupingSeparator is not None:
            literal = literal.replace(groupingSeparator, "")
        if decimalSeparator is not None and not (decimalSeparator == "."):
            literal = literal.replace(decimalSeparator, ".")
        return Decimal(literal, DefaultNumericType.DECIMAL128)

    def decimal(self, n: NUMBER, scale: NUMBER) -> NUMBER:
        if n is None or scale is None:
            return None

        return self.round(n, scale, NumericRoundingMode.HALF_EVEN)

    # Extension to DMN 1.3
    def round(self, n: NUMBER, scale: NUMBER, mode: NumericRoundingMode) -> NUMBER:
        if n is None or scale is None or mode is None or mode == NumericRoundingMode.UNKNOWN:
            return None

        prec = int(scale)
        if prec < 0:
            fmt = "1E{}".format(- prec)
        elif prec == 0:
            fmt = "1E0"
        else:
            fmt = "1E-{}".format(prec)
        return n.quantize(Decimal(fmt), rounding=mode.pMode)

    def floor(self, *args) -> NUMBER:
        # Extract n ans scale
        operands = varArgToList(*args)
        size = len(operands)
        if size == 0 or size > 2:
            return None
        elif size == 1:
            n = operands[0]
            scale = Decimal(0)
        else:
            n = operands[0]
            scale = operands[1]

        if n is None or scale is None:
            return None
        return self.round(n, scale, NumericRoundingMode.ROUND_FLOOR)

    def ceiling(self, *args) -> NUMBER:
        # Extract n ans scale
        operands = varArgToList(*args)
        size = len(operands)
        if size == 0 or size > 2:
            return None
        elif size == 1:
            n = operands[0]
            scale = Decimal(0)
        else:
            n = operands[0]
            scale = operands[1]

        if n is None or scale is None:
            return None
        return self.round(n, scale, NumericRoundingMode.ROUND_CEILING)

    def abs(self, n: NUMBER) -> NUMBER:
        if n is None:
            return None

        return n.copy_abs()

    def intModulo(self, dividend: NUMBER, divisor: NUMBER) -> NUMBER:
        if dividend is None or divisor is None:
            return None

        return Decimal(dividend % int(divisor))

    def modulo(self, dividend: NUMBER, divisor: NUMBER) -> NUMBER:
        if dividend is None or divisor is None:
            return None

        # dividend - divisor*floor(dividend/divisor)
        return dividend - (divisor * self.floor(dividend / divisor, Decimal(0)))

    def sqrt(self, number: NUMBER) -> NUMBER:
        if number is None:
            return None

        return number.sqrt(context=DefaultNumericType.DECIMAL128)

    def log(self, number: NUMBER) -> NUMBER:
        if number is None:
            return None

        return number.ln(context=DefaultNumericType.DECIMAL128)

    def exp(self, number: NUMBER) -> NUMBER:
        if number is None:
            return None

        return number.exp(context=DefaultNumericType.DECIMAL128)

    def odd(self, number: NUMBER) -> BOOLEAN:
        if not self.isIntegerValue(number):
            return None

        return int(number) % 2 != 0

    def even(self, number: NUMBER) -> BOOLEAN:
        if not self.isIntegerValue(number):
            return None

        return int(number) % 2 == 0

    #
    # List functions
    #
    def count(self, list_: List[Any]) -> NUMBER:
        if list_ is None:
            return Decimal(0)
        else:
            return Decimal(len(list_))

    def min(self, *operands) -> NUMBER:
        if len(operands) == 0:
            return None
        if len(operands) == 1:
            if isinstance(operands[0], list):
                operands = operands[0]
            else:
                operands = [operands[0]]

        result = operands[0]
        for opd in operands:
            if result > opd:
                result = opd
        return result

    def max(self, *operands) -> NUMBER:
        if len(operands) == 0:
            return None
        if len(operands) == 1:
            if isinstance(operands[0], list):
                operands = operands[0]
            else:
                operands = [operands[0]]

        result = operands[0]
        for opd in operands:
            if result < opd:
                result = opd
        return result

    def sum(self, *args) -> NUMBER:
        operands = varArgToList(*args)
        if len(operands) == 0:
            return None

        result = Decimal(0)
        for opd in operands:
            result += opd
        return result

    def mean(self, *args) -> NUMBER:
        operands = varArgToList(*args)
        if len(operands) == 0:
            return None

        sum = self.sum(operands)
        return DefaultNumericType.decimalNumericDivide(sum, Decimal(len(operands)))

    def product(self, *args) -> NUMBER:
        operands = varArgToList(*args)
        if len(operands) == 0:
            return None

        result = Decimal(1)
        for opd in operands:
            result = result * opd
        return result

    def median(self, *args) -> NUMBER:
        operands = varArgToList(*args)
        if len(operands) == 0:
            return None

        sortedList = sorted(operands, key=float)
        size = len(sortedList)
        if size % 2 == 0:
            first = sortedList[int(size / 2)]
            second = sortedList[int(size / 2 - 1)]
            median = (first + second) / 2
        else:
            median = sortedList[int(size / 2)]
        return median

    def stddev(self, *args) -> NUMBER:
        operands = varArgToList(*args)
        if len(operands) == 0:
            return None

        mean = self.mean(operands)
        length = Decimal(len(operands) - 1)
        variance = Decimal(0)
        for e in operands:
            number = e
            dm = number - mean
            dv = dm * dm
            variance += dv
        variance = DefaultNumericType.decimalNumericDivide(variance, length)
        stddev = self.sqrt(variance)
        return stddev

    def mode(self, *args) -> Optional[List[Decimal]]:
        if (args is None):
            return None
        operands = varArgToList(*args)
        if len(operands) == 0:
            return []

        max = -1
        modes = []
        countMap = defaultdict(int)
        for n in operands:
            if not isinstance(n, Decimal):
                return None
            countMap[n] += 1
            count = countMap[n]
            if count > max:
                max = count

        for key, value in countMap.items():
            if value == max:
                modes.append(key)

        sortedModes = sorted(modes)
        return sortedModes

    @staticmethod
    def toNumber(number: NUMBER) -> NUMBER:
        if isinstance(number, Decimal):
            return number
        else:
            return None

    @staticmethod
    def isIntegerValue(number):
        return number is not None and (number == int(number))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import List, Optional

from jdmn.runtime.DMNRuntimeException import DMNRuntimeException
from jdmn.runtime.RuleOutput import RuleOutput
from jdmn.runtime.annotation.HitPolicy import HitPolicy


class RuleOutputList:
    def __init__(self):
        self.matchedRuleResults = []

    def add(self, result: RuleOutput) -> None:
        if result is not None and result.isMatched():
            self.matchedRuleResults.append(result)

    def getMatchedRuleResults(self) -> List[RuleOutput]:
        return self.matchedRuleResults

    def noMatchedRules(self) -> bool:
        return len(self.matchedRuleResults) == 0

    def applySingle(self, hitPolicy: HitPolicy) -> Optional[RuleOutput]:
        if hitPolicy is None:
            hitPolicy = HitPolicy.UNIQUE
        if hitPolicy == HitPolicy.UNIQUE:
            if len(self.matchedRuleResults) == 1:
                return self.matchedRuleResults[0]
            else:
                return None
        elif hitPolicy == HitPolicy.ANY:
            if len(self.matchedRuleResults) > 0:
                distinctResults = set(self.matchedRuleResults)
                if len(distinctResults) == 1:
                    return next(iter(distinctResults))
                else:
                    return None
            else:
                return None
        elif hitPolicy == HitPolicy.FIRST:
            if len(self.matchedRuleResults) > 0:
                return self.matchedRuleResults[0]
            else:
                return None
        elif hitPolicy == HitPolicy.PRIORITY:
            if len(self.matchedRuleResults) > 0:
                sortedRuleOutputs = self.sort(self.matchedRuleResults)
                return sortedRuleOutputs[0]
            else:
                return None
        else:
            raise DMNRuntimeException("Not supported single hit policy '{}'.".format(hitPolicy.name))

    def applyMultiple(self, hitPolicy: HitPolicy) -> List[RuleOutput]:
        matchedRuleOutputs = self.getMatchedRuleResults()
        if hitPolicy == HitPolicy.COLLECT:
            return matchedRuleOutputs
        elif (hitPolicy == HitPolicy.RULE_ORDER):
            return matchedRuleOutputs
        elif (hitPolicy == HitPolicy.OUTPUT_ORDER):
            return self.sort(matchedRuleOutputs)
        else:
            raise DMNRuntimeException("Not supported multiple hit policy '{}'.".format(hitPolicy.name))

    def sort(self, matchedRuleOutputs: List[RuleOutput]) -> List[RuleOutput]:
        if len(matchedRuleOutputs) == 0:
            return matchedRuleOutputs
        else:
            result = matchedRuleOutputs[0]
            return result.sort(matchedRuleOutputs)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import date, time, datetime
from typing import Any

import isodate
from isodate import Duration, tzinfo

from jdmn.feel.lib.Types import STRING, DATE, TIME, DATE_TIME, DURATION, INT

DAY_NAMES = ["", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
MONTH_NAMES = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]


class DefaultDateTimeLib:
    #
    # Conversion functions
    #
    def date(self, *args) -> DATE:
        len_ = len(args)
        if len_ == 1:
            arg = args[0]
            if isinstance(arg, str):
                # From literal
                return date.fromisoformat(arg)
            elif isinstance(arg, datetime):
                # From datetime
                return arg.date()
            elif isinstance(arg, date):
                # From date
                return arg
            else:
                return None
        elif len_ == 3:
            # From year, month, day
            year = int(args[0])
            month = int(args[1])
            day = int(args[2])
            return date(year, month, day)
        else:
            return None

    # TODO ZoneIDs not supported in ISO 8601
    def time(self, *args) -> TIME:
        len_ = len(args)
        if len_ == 1:
            arg = args[0]
            if isinstance(arg, str):
                # From literal
                # return time.fromisoformat(literal)
                return isodate.parse_time(arg.upper())
            elif isinstance(arg, time):
                # From time
                return arg
            else:
                return None
        elif len_ == 4:
            # From hour, minute, second, offset
            hour = int(args[0])
            minute = int(args[1])
            seconds = int(args[2])
            return time(hour=hour, minute=minute, second=seconds, tzinfo=args[3])
        else:
            return None

    # TODO ZoneIDs not supported in ISO 8601
    def dateAndTime(self, *args) -> DATE_TIME:
        len_ = len(args)
        if len_ == 1:
            arg = args[0]
            if isinstance(arg, str):
                # From literal
                if "-" in arg and not ("T" in arg):
                    arg += "T00:00:00"
                return isodate.parse_datetime(arg.upper())
            elif isinstance(arg, time):
                # From time
                return arg
            else:
                return None
        elif len_ == 2:
            # From date, time
            return datetime.combine(args[0], args[1])
        else:
            return None

    #
    # Date properties
    #
    def year(self, date: DATE) -> INT:
        if date is None:
            return None

        return date.year

    def yearDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.year(dateTime)

    def month(self, date: DATE) -> INT:
        if date is None:
            return None

        return date.month

    def monthDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.month(dateTime)

    def day(self, date: DATE) -> INT:
        if date is None:
            return None

        return date.day

    def dayDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.day(dateTime)

    def weekday(self, date: DATE) -> INT:
        if date is None:
            return None

        return date.weekday() + 1

    def weekdayDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.weekday(dateTime)

    #
    # Time properties
    #
    def hour(self, time: TIME) -> INT:
        if time is None:
            return None

        return time.hour

    def hourDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.hourDateTime(dateTime)

    def minute(self, time: TIME) -> INT:
        if time is None:
            return None

        return time.minute

    def minuteDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.minute(dateTime)

    def second(self, time: TIME) -> INT:
        if time is None:
            return None

        return time.second

    def secondDateTime(self, dateTime: DATE_TIME) -> INT:
        return self.second(dateTime)

    def timeOffset(self, time: TIME) -> DURATION:
        if time is None:
            return None

        return self.toDuration(time.tzinfo)

    def timeOffsetDateTime(self, dateTime: DATE_TIME) -> DURATION:
        return self.timeOffset(dateTime)

    def timezone(self, time: TIME) -> STRING:
        if time is None:
            return None

        return self.toTimeZone(time.tzinfo)

    def timezoneDateTime(self, dateTime: DATE_TIME) -> STRING:
        return self.timezone(dateTime)

    #
    # Temporal functions
    #
    def dayOfYear(self, date: DATE) -> INT:
        if date is None:
            return None

        return date.timetuple().tm_yday

    def dayOfWeek(self, date: DATE) -> STRING:
        if date is None:
            return None

        dow: int = date.isocalendar()[2]
        return DAY_NAMES[dow]

    def weekOfYear(self, date: DATE) -> INT:
        if (date is None):
            return None

        return date.isocalendar()[1]

    def monthOfYear(self, date: DATE) -> STRING:
        if date is None:
            return None

        moy: int = date.month
        return MONTH_NAMES[moy - 1]

    #
    # Extra conversion functions
    #
    def toDate(self, from_: Any) -> DATE:
        if isinstance(from_, date):
            return from_
        elif isinstance(from_, datetime):
            return from_.date()
        else:
            return None

    def toTime(self, from_: Any) -> TIME:
        if isinstance(from_, time):
            return from_
        elif isinstance(from_, datetime):
            return from_.time()
        else:
            return None

    def toDateTime(self, from_: Any) -> DATE_TIME:
        if isinstance(from_, date):
            return datetime.combine(from_, time(0, 0, 0))
        elif isinstance(from_, datetime):
            return from_
        else:
            return None

    def toDuration(self, info: tzinfo) -> DURATION:
        if info is None:
            return None

        delta = info.utcoffset(None)
        seconds = int(delta.total_seconds())
        return Duration(seconds=seconds)

    def toTimeZone(self, info: tzinfo) -> STRING:
        if info is None:
            return None

        tzname = info.tzname(None)
        return tzname



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import logging

from jdmn.runtime.annotation.AnnotationSet import AnnotationSet
from jdmn.runtime.cache.Cache import Cache
from jdmn.runtime.cache.DefaultCache import DefaultCache
from jdmn.runtime.external.DefaultExternalFunctionExecutor import DefaultExternalFunctionExecutor
from jdmn.runtime.external.ExternalFunctionExecutor import ExternalFunctionExecutor
from jdmn.runtime.listener.EventListener import EventListener
from jdmn.runtime.listener.LoggingEventListener import LoggingEventListener


class ExecutionContext:
    LOGGER = logging.getLogger(__name__)

    def __init__(self, annotations: AnnotationSet = None,
                 eventListener: EventListener = None,
                 externalFunctionExecutor: ExternalFunctionExecutor = None,
                 cache: Cache = None):
        self.annotations = AnnotationSet() if annotations is None else annotations
        self.eventListener = LoggingEventListener(self.LOGGER) if eventListener is None else eventListener
        self.externalFunctionExecutor = DefaultExternalFunctionExecutor() if externalFunctionExecutor is None else externalFunctionExecutor
        self.cache = DefaultCache() if cache is None else cache



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any


class Pair:
    def __init__(self, left: Any, right: Any):
        self.left = left
        self.right = right

    def getLeft(self):
        return self.left

    def getRight(self):
        return self.right

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if isinstance(other, Pair):
            return False

        return self.left == other.left and self.right == other.right

    def __str__(self) -> str:
        return "Pair({0}, {1})".format(self.left, self.right)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import re
from datetime import datetime
from typing import Optional, Any

from lxml import etree
import elementpath

from jdmn.feel.lib.Types import STRING, BOOLEAN, INT, LIST, NUMBER


class DefaultStringLib:
    def string(self, from_: Any) -> STRING:
        if isinstance(from_, datetime):
            return str(from_.isoformat())

        return str(from_)

    def contains(self, string: STRING, match: STRING) -> BOOLEAN:
        if string is None or match is None:
            return None

        return match in string

    def startsWith(self, string: Optional[str], match: Optional[str]) -> Optional[bool]:
        if string is None or match is None:
            return None

        return string.startswith(match)

    def endsWith(self, string: STRING, match: STRING) -> BOOLEAN:
        if string is None or match is None:
            return None

        return string.endswith(match)

    def stringLength(self, string: STRING) -> INT:
        if string is None:
            return None

        # The number of Unicode code units in the string
        bytes = string.encode('utf-16', 'surrogatepass')
        # The number of characters (Unicode code point)
        transformed = bytes.decode('utf-16')
        result = len(transformed)
        return result

    def substring(self, string: STRING, startPosition: NUMBER, length: NUMBER) -> STRING:
        if string is None or startPosition is None:
            return None

        start = int(startPosition)
        if start < 0:
            start = len(string) + start
        else:
            start -= 1

        normal = self.normalizeSurrogatePairs(string)
        end = len(normal) if length is None else start + int(length)
        result = self.substringCodePoints(normal, start, end)
        return result

    def upperCase(self, string: STRING) -> STRING:
        if string is None:
            return None

        return string.upper()

    def lowerCase(self, string: STRING) -> STRING:
        if string is None:
            return None

        return string.lower()

    def substringBefore(self, string: STRING, match: STRING) -> STRING:
        if string is None or match is None:
            return None

        i = string.find(match)
        return "" if i == -1 else string[0:i]

    def substringAfter(self, string: STRING, match: STRING) -> STRING:
        if string is None or match is None:
            return None

        i = string.find(match)
        return "" if i == -1 else string[i + len(match):]

    def replace(self, input: STRING, pattern: STRING, replacement: STRING, flags: STRING) -> STRING:
        if input is None or pattern is None or replacement is None:
            return None
        if flags is None:
            flags = ""

        expression = "replace(/root, '{}', '{}', '{}')".format(pattern, replacement, flags)
        return self.evaluateXPath(input, expression)

    def matches(self, input: STRING, pattern: STRING, flags: STRING) -> BOOLEAN:
        if input is None or pattern is None:
            return None
        if flags is None:
            flags = ""

        expression = "/root[matches(., '{}', '{}')]".format(pattern, flags)
        value = self.evaluateXPath(input, expression)
        return len(value) != 0

    def split(self, string: STRING, delimiter: STRING) -> LIST:
        if string is None or delimiter is None:
            return None
        if string.strip() == "" or delimiter.strip() == "":
            return None

        result = []
        pattern = re.compile(delimiter)
        start = 0
        for match in re.finditer(pattern, string):
            delimiterStart = match.start()
            delimiterEnd = match.end()
            token = string[start:delimiterStart]
            result.append(token)
            start = delimiterEnd
        if start <= len(string):
            token = string[start:]
            result.append(token)
        return result

    @staticmethod
    def evaluateXPath(input, expression):
        # Read document
        xmlStr = "<root>" + input + "</root>"
        root = etree.fromstring(xmlStr)

        # Evaluate xpath
        res1 = elementpath.select(root, expression)
        return res1

    @staticmethod
    def normalizeSurrogatePairs(string):
        if string is None:
            return None

        # Roundtrip to utf-16 to replace surrogate pairs
        utf16bytes = string.encode('utf-16', 'surrogatepass')
        utf16chars = utf16bytes.decode('utf-16')
        utf8 = utf16chars.encode('utf-8').decode('utf-8')
        return utf8

    @staticmethod
    def substringCodePoints(string, start: int, end: int) -> str:
        result = ""
        cps = [ord(c) for c in string]
        for i, v in enumerate(cps):
            if start <= i < end:
                result += chr(v)
        return result
